@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2020 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -38,7 +38,6 @@
 
 #ifdef WIN32
  #include <winsock.h>
-
 #else /* UNIX */
  #include <sys/types.h>
  #include <time.h>
@@ -55,6 +54,7 @@
 #include <sys/timeb.h>
 
 #include "property.h"
+#include "wrapper_jvminfo.h"
 
 #ifndef WIN32
  /*
@@ -66,6 +66,10 @@
  #endif
 
 extern char** environ;
+#else
+ #if !defined(WINIA) && defined(SERVICE_ACCEPT_PRESHUTDOWN)
+  #define SUPPORT_PRESHUTDOWN
+ #endif
 #endif
 
 /* The following define will enable debug output of the code to parse the JVM output. */
@@ -182,6 +186,16 @@ extern char** environ;
 #define WRAPPER_ACTION_SOURCE_CODE_SIGNAL                  5  /* Action originated from a signal. */
 #define WRAPPER_ACTION_SOURCE_CODE_PING_TIMEOUT            11 /* Action originated from a timeout. */
 
+/* Required services access rights per action
+ * ATTENTION: Any change in the definitions below should be reported to the functions & variables in wrapper_service_permission_win.c  */
+#ifdef WIN32
+ #define WRAPPER_SERVICE_QUERY_STATUS        (SERVICE_QUERY_STATUS | SERVICE_QUERY_CONFIG)
+ #define WRAPPER_SERVICE_START               (SERVICE_QUERY_STATUS | SERVICE_START)
+ #define WRAPPER_SERVICE_STOP                (SERVICE_QUERY_STATUS | SERVICE_STOP)
+ #define WRAPPER_SERVICE_PAUSE_CONTINUE      (SERVICE_QUERY_STATUS | SERVICE_PAUSE_CONTINUE)
+ #define WRAPPER_SERVICE_CONTROL_CODE        (SERVICE_QUERY_STATUS | SERVICE_USER_DEFINED_CONTROL)
+#endif
+
 /* Because of the way time is converted to ticks, the largest possible timeout that
  *  can be specified without causing 32-bit overflows is (2^31 / 1000) - 5 = 2147478
  *  Which is a little over 24 days.  To make the interface nice, round this down to
@@ -202,6 +216,13 @@ typedef unsigned int TICKS;
 typedef unsigned long TICKS;
 #endif
 
+#define STATE_OUTPUT_MODE_DEFAULT   1
+#define STATE_OUTPUT_MODE_SECONDS   2
+#define STATE_OUTPUT_MODE_CHANGED   4
+
+#define STATE_OUTPUT_MASK_ALL       (STATE_OUTPUT_MODE_DEFAULT | STATE_OUTPUT_MODE_SECONDS | STATE_OUTPUT_MODE_CHANGED)
+#define STATE_OUTPUT_MASK_NOSECS    (STATE_OUTPUT_MASK_ALL & (~STATE_OUTPUT_MODE_SECONDS))
+
 #ifdef WIN32
 /* Defines the maximum number of service manager control events that can be queued in a single loop. */
 #define CTRL_CODE_QUEUE_SIZE 26 /* Can enqueue one less than this count at any time. */
@@ -217,7 +238,6 @@ typedef unsigned long TICKS;
 typedef struct PendingPing PendingPing, *PPendingPing;
 struct PendingPing {
     TICKS sentTicks;
-    TICKS timeoutTicks;
     TICKS slowTicks;
     PPendingPing nextPendingPing;
 };
@@ -225,7 +245,7 @@ struct PendingPing {
 /* Type definitions */
 typedef struct WrapperConfig WrapperConfig;
 struct WrapperConfig {
-    TCHAR   *argBinary;             /* The name of the wrapper binary. */
+    TCHAR   *argBinary;             /* The path to the wrapper binary. */
     TCHAR   *argCommand;            /* The command used to launch the wrapper. */
     TCHAR   *argCommandArg;         /* The argument to the command used to launch the wrapper. */
     TCHAR   *argConfFile;           /* The name of the config file from the command line. */
@@ -238,7 +258,11 @@ struct WrapperConfig {
     int     javaArgValueCount;      /* Number of the arguments getting passed over to the java application */
 
     TCHAR   *initialPath;           /* What the working directory was when the Wrapper process was first launched. */
-    TCHAR   *language;              /* The language */
+    TCHAR   *baseName;              /* The name of the wrapper binary without the OS, arch and bits. */
+#ifdef WIN32
+    UINT    jvm_stdout_codepage;    /* The code page used for JVM outputs. */
+#endif
+    int     use_sun_encoding;       /* TRUE if the Wrapper uses the value of sun.stdout.encoding to read JVM output. */
     int     backendType;            /* The type of the backend that the Wrapper and Java use to communicate. */
     int     configured;             /* TRUE if loadConfiguration has been called. */
     int     useSystemTime;          /* TRUE if the wrapper should use the system clock for timing, FALSE if a tick counter should be used. */
@@ -263,7 +287,15 @@ struct WrapperConfig {
     TCHAR   *configFile;            /* Name of the configuration file */
     int     commandLogLevel;        /* The log level to use when logging the java command. */
     int     printJVMVersion;        /* tells the Wrapper to create a temp JVM to query the version, before starting the java application */
+    JavaVersion *javaVersion;       /* Java version of the current or next JVM instance to be launched. */
+    JavaVersion *javaVersionMin;    /* The minimum version of Java required for the Wrapper to launch a JVM. */
+    JavaVersion *javaVersionMax;    /* The maximum version of Java required for the Wrapper to launch a JVM. */
+    int     jvmBits;                /* JVM bits of the current or next JVM instance to be launched. */
+    int     jvmVendor;              /* JVM implementation (Oracle, IBM, etc.) of the current or next JVM instance to be launched. */
+    int     jvmDefaultLogLevel;     /* The default log level used for JVM outputs. */
+    int     jvmSource;              /* The source used for JVM outputs. */
 #ifdef WIN32
+    TCHAR   *registry_java_home;    /* Path to the Java Home when the command is located to the registry, or NULL otherwise. */
     TCHAR   *jvmVersionCommand;     /* Command used to launch the JVM and request its version */
     TCHAR   *jvmCommand;            /* Command used to launch the JVM */
 #else /* UNIX */
@@ -317,12 +349,16 @@ struct WrapperConfig {
     int     jStateTimeoutTicksSet;  /* 1 if the current jStateTimeoutTicks is set. */
     TICKS   lastPingTicks;          /* Time that the last ping was sent */
     TICKS   lastLoggedPingTicks;    /* Time that the last logged ping was sent */
-
+    int     environmentLogLevel;    /* Log Level at which the environment variables should be logged. */
     int     isDebugging;            /* TRUE if set in the configuration file */
     int     isAdviserEnabled;       /* TRUE if advice messages should be output. */
     const TCHAR *nativeLibrary;     /* The base name of the native library loaded by the WrapperManager. */
     int     libraryPathAppendPath;  /* TRUE if the PATH environment variable should be appended to the java library path. */
     int     isStateOutputEnabled;   /* TRUE if set in the configuration file.  Shows output on the state of the state engine. */
+    int     stateOutputMode;        /* Mode used to print the state output. */
+#ifdef WIN32
+    int     isMessageOutputEnabled; /* TRUE if detailed Message output should be included in debug output. */
+#endif
     int     isJavaIOOutputEnabled;  /* TRUE if detailed javaIO output should be included in debug output. */
     int     isTickOutputEnabled;    /* TRUE if detailed tick timer output should be included in debug output. */
     int     isLoopOutputEnabled;    /* TRUE if very detailed output from the main loop should be output. */
@@ -345,6 +381,7 @@ struct WrapperConfig {
     int     isTestsDisabled;        /* TRUE if the use of tests in the WrapperManager class should be disabled. */
     int     isShutdownHookDisabled; /* TRUE if the use of a shutdown hook by the WrapperManager class should be disabled. */
     int     isForcedShutdownDisabled; /* TRUE if forced shutdowns are disabled. */
+    int     forcedShutdownDelay;    /* minimum amount of time required between two CTRL-C or TERM signals. */
     int     startupDelayConsole;    /* Delay in seconds before starting the first JVM in console mode. */
     int     startupDelayService;    /* Delay in seconds before starting the first JVM in service mode. */
     int     exitCode;               /* Code which the wrapper will exit with */
@@ -362,6 +399,7 @@ struct WrapperConfig {
     int     restartReloadConf;      /* TRUE if the configuration should be reloaded before a JVM restart. */
     int     isRestartDisabled;      /* TRUE if restarts should be disabled. */
     int     isAutoRestartDisabled;  /* TRUE if automatic restarts should be disabled. */
+    int     runWithoutJVM;          /* TRUE if the Wrapper should run without launching a JVM (used for testing). */
     int     requestThreadDumpOnFailedJVMExit; /* TRUE if the JVM should be asked to dump its state when it fails to halt on request. */
     int     requestThreadDumpOnFailedJVMExitDelay; /* Number of seconds to wait after the thread dump before killing the JVM. */
     TICKS   jvmLaunchTicks;         /* The tick count at which the previous or current JVM was launched. */
@@ -397,6 +435,18 @@ struct WrapperConfig {
     int     statusFileUmask;        /* Umask to use when creating the status file. */
     int     javaStatusFileUmask;    /* Umask to use when creating the java status file. */
     int     anchorFileUmask;        /* Umask to use when creating the anchor file. */
+#ifndef WIN32
+    int     groupStrict;            /* Whether the Wrapper should stop or not when it detects invalid values for the wrapper.*.group properties while loading the configuration. */
+    gid_t   group;                  /* Default group for all files (Unlike umask, I don't think there is a way to automatically specify a default group for each file created by the process because the underlying function needs the filepath. So this member only specifies the default value for the files below). */
+ /* gid_t   javaGroup; */           /* Default group for the java process. *//* TODO: can this exist? */
+    gid_t   pidFileGroup;           /* Group to use when creating the pid file. */
+    gid_t   lockFileGroup;          /* Group to use when creating the lock file. */
+    gid_t   javaPidFileGroup;       /* Group to use when creating the java pid file. */
+    gid_t   javaIdFileGroup;        /* Group to use when creating the java id file. */
+    gid_t   statusFileGroup;        /* Group to use when creating the status file. */
+    gid_t   javaStatusFileGroup;    /* Group to use when creating the java status file. */
+    gid_t   anchorFileGroup;        /* Group to use when creating the anchor file. */
+#endif
     int     ignoreSignals;          /* Mask that determines where the Wrapper should ignore any catchable system signals.  Can be ingored in the Wrapper and/or JVM. */
     TCHAR   *consoleTitle;          /* Text to set the console title to. */
     TCHAR   *serviceName;           /* Name of the service. */
@@ -423,7 +473,7 @@ struct WrapperConfig {
     int     ntServicePasswordPromptMask; /* If true then the password will be masked as it is input. */
     int     ntServiceInteractive;   /* Should the service be allowed to interact with the desktop? */
     int     ntHideJVMConsole;       /* Should the JVMs Console window be hidden when run as a service.  True by default but GUIs will not be visible for JVMs prior to 1.4.0. */
-    int     ntHideWrapperConsole;   /* Should the Wrapper Console window be hidden when run as a service. */
+    int     ntShowWrapperConsole;   /* Should the Wrapper Console window be shown when run as a service/wrapperw. */
     HINSTANCE wrapperHInstance;     /* The HINSTANCE of the Wrapper process. */
     int     wrapperConsoleHide;     /* True if the Wrapper Console window should be hidden. */
     HWND    wrapperConsoleHWND;     /* The HWND of the Wrapper's console if it was located. */
@@ -432,7 +482,7 @@ struct WrapperConfig {
     int     jvmConsoleVisible;      /* True if the JVM Console window is visible. */
     int     ntAllocConsole;         /* True if a console should be allocated for the Service. */
     int     generateConsole;        /* Make sure that a console is always generated to support thread dumps */
-#ifndef WINIA
+#ifdef SUPPORT_PRESHUTDOWN
     int     ntPreshutdown;          /* True if the service accepts SERVICE_CONTROL_PRESHUTDOWN controle code, False if it accepts SERVICE_CONTROL_SHUTDOWN. */
     int     ntPreshutdownTimeout;   /* Number of seconds that the service controler will wait after sending the SERVICE_CONTROL_PRESHUTDOWN notification. */
 #endif
@@ -470,7 +520,7 @@ struct WrapperConfig {
     int     ctrlCodeContinueTrapped;/* SERVICE_CONTROL_CONTINUE was trapped. */
     int     ctrlCodeStopTrapped;    /* SERVICE_CONTROL_STOP was trapped. */
     int     ctrlCodeShutdownTrapped;/* SERVICE_CONTROL_SHUTDOWN was trapped. */
- #ifndef WINIA
+ #ifdef SUPPORT_PRESHUTDOWN
     int     ctrlCodePreShutdownTrapped;/* SERVICE_CONTROL_PRESHUTDOWN was trapped. */
  #endif
     int     ctrlCodeDumpTrapped;    /* The configured thread dump control code was trapped. */
@@ -516,18 +566,43 @@ struct WrapperConfig {
 #define WRAPPER_MSG_PAUSE         (char)138
 #define WRAPPER_MSG_RESUME        (char)139
 #define WRAPPER_MSG_GC            (char)140
+#ifdef WIN32
+ #define WRAPPER_MSG_FIRE_CTRL_EVENT (char)143
+#endif
 
 #define WRAPPER_PROCESS_DOWN      200
 #define WRAPPER_PROCESS_UP        201
 
-/* default timeout (in second) when executing child process to print java version */
-#define DEFAULT_JAVA_VERSION_TIMEOUT    10
+/* default timeout (in second) when executing child process to print java version (in 3.5.42 the value was increased from 10 to 30) */
+#define DEFAULT_JAVA_VERSION_TIMEOUT    30
 
 extern WrapperConfig *wrapperData;
 extern Properties    *properties;
 
 extern TCHAR wrapperClasspathSeparator;
 
+/**
+ * Dumps the table of environment variables, and their sources.
+ */
+void dumpEnvironment();
+
+/**
+ * Return TRUE if the this is a prompt call made from the script (like --translate or --jvm_bits).
+ *
+ * @param argCommand the first arguement passed when launching the Wrapper
+ */
+int isPromptCallCommand(const TCHAR* argCommand);
+
+/**
+ * Return TRUE if the this is a prompt call made from the script (like -translate or -jvm_bits).
+ */
+int isPromptCall();
+
+/**
+ * Attempt to set the console title if it exists and is accessible.
+ */
+extern void wrapperSetConsoleTitle();
+
 /* Protocol Functions */
 /**
  * Close the backend socket.
@@ -658,6 +733,14 @@ extern int initializeTimer();
 extern void disposeTimer();
 
 extern int showHostIds(int logLevel);
+#ifdef WIN32
+/**
+ * Returns TRUE if the Wrapper process is associated with a console or if one
+ *  will be allocated a later stage. This function should be called after the
+ *  configuration file has been read.
+ */
+int wrapperProcessHasVisibleConsole();
+#endif
 extern void wrapperLoadHostName();
 
 /**
@@ -685,18 +768,20 @@ extern int *wrapperGetActionListForNames(const TCHAR *actionNameList, const TCHA
  */
 extern void wrapperProcessActionList(int *actionList, const TCHAR *triggerMsg, int actionSourceCode, int actionPropertyIndex, int logForActionNone, int exitCode);
 
-extern void wrapperAddDefaultProperties();
+extern void wrapperAddDefaultProperties(Properties *props);
+
+extern int getOriginalWorkingDir();
 
 extern int wrapperLoadConfigurationProperties(int preload);
 
-extern void wrapperGetCurrentTime(struct timeb *timeBuffer);
-
 extern void updateStringValue(TCHAR **ptr, const TCHAR *value);
 
 #ifdef WIN32
+int confReadFilterCallbackMB(const char *bufferMB);
 
 extern void wrapperInitializeProfileCounters();
 extern void wrapperDumpPageFaultUsage();
+extern void disposeSystemPath();
 extern TCHAR** wrapperGetSystemPath();
 extern int wrapperGetJavaHomeFromWindowsRegistry(TCHAR *javaHome);
 #endif
@@ -709,20 +794,11 @@ extern int wrapperBuildJavaClasspath(TCHAR **classpath);
  * length is the number of strings in the above array.
  */
 extern int wrapperBuildJavaCommandArray(TCHAR ***strings, int *length, int addQuotes, const TCHAR *classpath);
-extern void wrapperFreeJavaCommandArray(TCHAR **strings, int length);
+extern void wrapperFreeStringArray(TCHAR **strings, int length);
 
-extern int wrapperInitialize();
+extern int wrapperInitialize(int silent);
 extern void wrapperDispose();
 
-/**
- * Returns the file name base as a newly malloced TCHAR *.  The resulting
- *  base file name will have any path and extension stripped.
- *
- * baseName should be long enough to always contain the base name.
- *  (strlen(fileName) + 1) is safe.
- */
-extern void wrapperGetFileBase(const TCHAR *fileName, TCHAR *baseName);
-
 /**
  * Output the version.
  */
@@ -747,6 +823,20 @@ extern int wrapperParseArguments(int argc, TCHAR **argv);
  */
 extern void wrapperJVMProcessExited(TICKS nowTicks, int exitCode);
 
+/**
+ * Comfirm that the Java version is in the range in which the Wrapper is allowed to run.
+ *
+ * @return TRUE if the Java version is ok, FALSE otherwise.
+ */
+extern int wrapperConfirmJavaVersion();
+
+/**
+ * Set the Java version that the Wrapper will use.
+ *
+ * @param javaVersion The Java version to set or NULL if a default value should be set.
+ */
+void wrapperSetJavaVersion(JavaVersion* javaVersion);
+
 /**
  * Read and process any output from the child JVM Process.
  *
@@ -764,6 +854,11 @@ extern void wrapperJVMProcessExited(TICKS nowTicks, int exitCode);
  */
 extern int wrapperReadChildOutput(int maxTimeMS);
 
+/**
+ * Read the output returned by the 'java -version' command.
+ */
+extern void wrapperReadJavaVersionOutput();
+
 /**
  * Changes the current Wrapper state.
  *
@@ -804,6 +899,7 @@ extern void wrapperCheckConsoleWindows();
 extern void verifyEmbeddedSignature();
 
 extern int exceptionFilterFunction(PEXCEPTION_POINTERS exceptionPointers);
+BOOL isSecondary();
 BOOL extern elevateThis(int argc, TCHAR **argv);
 BOOL extern duplicateSTD();
 BOOL extern myShellExec(HWND hwnd, LPCTSTR pszVerb, LPCTSTR pszPath, LPCTSTR pszParameters, LPCTSTR pszDirectory, TCHAR* namedPipeName);
@@ -813,10 +909,12 @@ BOOL extern isWin10OrHigher();
 BOOL extern isVista();
 BOOL extern isWinXP();
 extern void wrapperMaintainControlCodes();
-int wrapperServiceStatus(int consoleOutput);
+int wrapperServiceStatus(const TCHAR* serviceName, const TCHAR* serviceDisplayName, int consoleOutput);
+extern int wrapperRemove();
 #define OSBUFSIZE 256
 BOOL GetOSDisplayString(TCHAR** pszOS);
 #else
+extern void changePidFileGroup(const TCHAR* filename, gid_t newGroup);
 extern void wrapperMaintainSignals();
 extern TCHAR* findPathOf(const TCHAR *exe, const TCHAR* name);
 #endif
@@ -869,12 +967,24 @@ extern int wrapperGetProcessStatus(TICKS nowTicks, int sigChild);
  */
 extern void wrapperPauseBeforeExecute();
 
+/**
+ * Create a child process to print the Java version running the command:
+ *    /path/to/java -version
+ *  After printing the java version, the process is terminated.
+ * 
+ * In case the JVM is slow to start, it will time out after
+ * the number of seconds set in "wrapper.java.version.timeout".
+ * 
+ * Note: before the timeout is reached, the user can ctrl+c to stop the Wrapper.
+ */
+extern int wrapperLaunchJavaVersion();
+
 /**
  * Launches a JVM process and store it internally
  *
  * @return TRUE if there were any problems.  When this happens the Wrapper will not try to restart.
  */
-extern int wrapperExecute();
+extern int wrapperLaunchJavaApp();
 
 /**
  * Returns a tick count that can be used in combination with the
@@ -1024,6 +1134,8 @@ extern size_t wrapperQuoteValue(const TCHAR* value, TCHAR *buffer, size_t buffer
  */
 extern int wrapperCheckQuotes(const TCHAR *value, const TCHAR *propName);
 
+extern int getStateOutputModeForName(const TCHAR *name);
+
 /**
  * The main event loop for the wrapper.  Handles all state changes and events.
  */
@@ -1036,6 +1148,20 @@ extern void wrapperBuildKey();
  */
 extern void wrapperRequestDumpJVMState();
 
+/**
+ * Build the command line used to get the Java version.
+ *
+ * @return TRUE if there were any problems.
+ */
+extern int wrapperBuildJavaVersionCommand();
+
+/**
+ * Builds up the java command section of the Java command line.
+ *
+ * @return The final index into the strings array, or -1 if there were any problems.
+ */
+extern int wrapperBuildJavaCommandArrayJavaCommand(TCHAR **strings, int addQuotes, int detectDebugJVM, int index);
+
 /**
  * Build the java command line.
  *
@@ -1131,6 +1257,18 @@ extern void wrapperStoppedSignaled();
 extern void wrapperStartPendingSignaled(int waitHint);
 extern void wrapperStartedSignaled();
 
+#ifdef WIN32
+int wrapperCheckPPid(DWORD pid, DWORD ppid);
+#endif
+
+int checkPidFile();
+int writePidFile(const TCHAR *filename, DWORD pid, int newUmask
+#ifndef WIN32
+            , gid_t newGroup
+#endif
+            );
+int wrapperWriteStartupPidFiles();
+
 /******************************************************************************
  * Inner types and methods for loading Wrapper configuration.
  *****************************************************************************/
