@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2019 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2020 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -93,7 +93,6 @@ int pipedes[2] = {-1, -1};
 
 TCHAR wrapperClasspathSeparator = TEXT(':');
 
-int javaIOThreadSet = FALSE;
 pthread_t javaIOThreadId;
 int javaIOThreadStarted = FALSE;
 int stopJavaIOThread = FALSE;
@@ -800,14 +799,11 @@ int initializeJavaIO() {
     if (res) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
             TEXT("Unable to create a javaIO thread: %d, %s"), res, getLastErrorText());
-        javaIOThreadSet = TRUE;
         return 1;
     } else {
         if (pthread_detach(javaIOThreadId)) {
-            javaIOThreadSet = TRUE;
             return 1;
         }
-        javaIOThreadSet = FALSE;
         return 0;
     }
 }
@@ -1014,17 +1010,6 @@ int wrapperInitializeRun() {
             return res;
         }
     }
-    
-    if (wrapperData->useJavaIOThread) {
-        /* Create and initialize a javaIO thread. */
-        if ((res = initializeJavaIO()) != 0) {
-            return res;
-        }
-    } else {
-        javaIOThreadSet = FALSE;
-        /* Unable to set the javaIOThreadId to a null value on all platforms
-         * javaIOThreadId = 0;*/
-    }
 
     return retval;
 }
@@ -1299,7 +1284,7 @@ size_t wrapperCalculateEnvironmentLength() {
  *
  * @return TRUE if there were any problems, FALSE otherwise.
  */
-int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
+int wrapperLaunchJvm(TCHAR** command, pid_t *ppid) {
     int i;
     pid_t proc;
     int execErrno;
@@ -1308,7 +1293,7 @@ int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
 
     /* Create the pipe. */
     if (pipe(pipedes) < 0) {
-        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                    TEXT("Could not init pipe: %s"), getLastErrorText());
         return TRUE;
     }
@@ -1329,7 +1314,7 @@ int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
         /* Restore the auto close flag. */
         setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);
 
-        log_printf(WRAPPER_SOURCE_WRAPPER, errorLevel,
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                    TEXT("Could not spawn JVM process: %s"), getLastErrorText());
 
         /* The fork failed so there is no child side.  Close the pipes so we don't attempt to read them later. */
@@ -1396,7 +1381,7 @@ int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
             /* TODO: Figure out a way to inform the Wrapper not to restart and try again as repeatedly doing this is meaningless. */
         }
 
-        if (wrapperData->isAdviserEnabled && (errorLevel == LEVEL_FATAL)) {
+        if (wrapperData->isAdviserEnabled) {
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("%s"), LOG_FORK_MARKER );
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE,
                 TEXT("%s------------------------------------------------------------------------"), LOG_FORK_MARKER );
@@ -1418,7 +1403,12 @@ int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
         }
 
         /* This process needs to end. */
-        exit(wrapperData->errorExitCode);
+        if (command == wrapperData->jvmVersionCommand) {
+            /* Set an exit code to distinguish cases where we failed to execute the java command. */
+            exit(127);
+        } else {
+            exit(wrapperData->errorExitCode);
+        }
         return TRUE; /* Will not get here. */
     } else {
         /* We are the parent side and need to assume that at this point the JVM is up. */
@@ -1456,8 +1446,6 @@ int wrapperLaunchJvm(TCHAR** command, pid_t *ppid, int errorLevel) {
  * 
  * In case the JVM is slow to start, it will time out after
  * the number of seconds set in "wrapper.java.version.timeout".
- * 
- * Note: before the timeout is reached, the user can ctrl+c to stop the Wrapper.
  */
 int wrapperLaunchJavaVersion() {
     int i;
@@ -1479,14 +1467,14 @@ int wrapperLaunchJavaVersion() {
      * (we know this this JVM is launched without system properties specifying a different encoding). */
     resetJvmOutputEncoding(FALSE);
     
-    if (wrapperLaunchJvm(wrapperData->jvmVersionCommand, &pid, LEVEL_ERROR)) {
+    if (wrapperLaunchJvm(wrapperData->jvmVersionCommand, &pid)) {
         return TRUE;
     }
     
     /* Note: on CentOS, in case we don't call waitpid(), then the child process becomes a zombie */
-    /* Note2: use _tprintf to log messages instead of log_printf. With log_printf, the columns are printed twice */
 
-    /* If the user set the value to 0, then we will wait indefinitely. */
+    /* If the user set the value to 0, then we will wait indefinitely.
+     *  NOTE: it would be better to create a new state in the main event loop as this would allow the user to press CTRL-C. */
     blockTimeout = getIntProperty(properties, TEXT("wrapper.java.version.timeout"), DEFAULT_JAVA_VERSION_TIMEOUT) * 1000;
     
     if (blockTimeout > 0) {
@@ -1501,24 +1489,42 @@ int wrapperLaunchJavaVersion() {
         /* Wait indefinately. */
         result = waitpid(pid, &status, 0);
     }
+
+    if (result > 0) {
+        /* Process completed - we know that nothing more can be written to stdout/stderr. */
+        if (WIFEXITED(status) && (WEXITSTATUS(status) == 127)) {
+            /* Read all messages written in the "child" code of the fork.
+             *  NOTES:
+             *   - the messages will be formatted as if they were printed by the Wrapper because of the LOG_FORK_MARKER prefix.
+             *   - there will be no attempt to parse the Java version because the source is not set to WRAPPER_SOURCE_JVM_VERSION. */
+            while (wrapperReadChildOutput(250)) { }
             
-    if (result == 0) {
-        /* Timed out. */
-        _tprintf(TEXT("Child process: Java version: timed out\n"));
-        kill(pid, SIGKILL);
-    } else if (result > 0) {
-        /* Process completed. */
-#ifdef _DEBUG
-        _tprintf(TEXT("Child process: Java version: successful\n"));
-#endif
+            /* We called exit(127) in the child process, which means the java command could not be executed.
+             *  We need to stop as we will not be able to launch any other JVMs...
+             *  This will behave the same as on Windows where wrapperLaunchJvm() returns TRUE. */
+            return TRUE;
+        }
+
+        wrapperReadJavaVersionOutput();
     } else {
-        /* Wait failed. */
-        _tprintf(TEXT("Child process: Java version: wait failed\n"));
-        kill(pid, SIGKILL);
+        if (result == 0) {
+            /* Timed out. */
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Child process: Java version: timed out"));
+        } else {
+            /* Wait failed. */
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Child process: Java version: wait failed"));
+        }
+        if (kill(pid, SIGKILL) != 0) {
+            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Child process: Java version: kill failed - %s"), getLastErrorText());
+            
+            /* Don't read the output but resolve the Java version to its default value. */
+            wrapperSetJavaVersion(NULL);
+        } else {
+            /* The process is now killed. There might be no output but read all the pipe anyway.
+             *  If the Java version can't be found, it will be resolved to the default value. */
+            wrapperReadJavaVersionOutput();
+        }
     }
-
-    wrapperReadJavaVersionOutput();
-
     return FALSE;
 }
 
@@ -1528,6 +1534,7 @@ int wrapperLaunchJavaVersion() {
  * @return TRUE if there were any problems.  When this happens the Wrapper will not try to restart.
  */
 int wrapperLaunchJavaApp() {
+    static int javaIOThreadSet = FALSE;
     int i;
     pid_t pid;
 
@@ -1557,8 +1564,18 @@ int wrapperLaunchJavaApp() {
         return TRUE;
     }
     
+    if (wrapperData->useJavaIOThread) {
+        /* Create and initialize a javaIO thread. */
+        if (!javaIOThreadSet) {
+            if (initializeJavaIO()) {
+                return TRUE;
+            }
+            javaIOThreadSet = TRUE;
+        }
+    }
+    
     /* Now launch the JVM process. */
-    if (wrapperLaunchJvm(wrapperData->jvmCommand, &pid, LEVEL_FATAL)) {
+    if (wrapperLaunchJvm(wrapperData->jvmCommand, &pid)) {
         wrapperData->exitCode = wrapperData->errorExitCode;
         return TRUE;
     }
@@ -1795,7 +1812,7 @@ int wrapperReadChildOutputBlock(char *blockBuffer, int blockSize, int *readCount
         return FALSE;
     }
 
-#if defined OPENBSD || defined FREEBSD
+#ifdef FREEBSD
     /* Work around FreeBSD Bug #kern/64313
      *  http://www.freebsd.org/cgi/query-pr.cgi?pr=kern/64313
      *
